# from dh_potluck.email.email_service import EmailService
#
# # EmailService.send_email_html(
# #     [
# #         {
# #             'email': 'jake.delmas@dashhudson.com',
# #             'name': 'jake delmas',
# #             'type': 'to'
# #         },
# #         {
# #             'email': 'jakemdelmas@gmail.com',
# #             'name': 'jake delmas',
# #             'type': 'cc'
# #         }
# #     ],
# #     subject='Test Subject',
# #     html='<b>this is a test</b>',
# # )
# # EmailService.send_email_template(
# #     [
# #         {
# #             'email': 'jake.delmas@dashhudson.com',
# #             'name': 'jake delmas',
# #             'type': 'to'
# #         },
# #         {
# #             'email': 'jakemdelmas@gmail.com',
# #             'name': 'jake delmas',
# #             'type': 'cc'
# #         }
# #     ],
# #     'restore-password',
# #     [{'name': 'link', 'content': 'test'}, {'name': 'name', 'content': 'JAKEEE'}],
# #     subject='template test',
# # )
#
# test = 'dddd' 'ccccc'
# print(test)



import importlib
import logging
import os
import threading
import traceback
from http import HTTPStatus

import pkg_resources
from ddtrace import tracer
from flask import Blueprint, g, jsonify, render_template, request
from flask_limiter import Limiter
from marshmallow import ValidationError
from sqlalchemy.exc import IntegrityError
from werkzeug.exceptions import HTTPException, InternalServerError
from werkzeug.middleware.proxy_fix import ProxyFix

from .auth import get_user, role_required
from .logging import add_request_params_to_trace, patch_celery_get_logger, structure_logger
from .platform_connection import PlatformConnectionError
from .queries_summary import get_database_queries_summary


class DHPotluck:
    _app = None
    _app_token = None
    _structured_logging_enabled = False
    _rate_limiting_enabled = False
    _rate_limit = None
    _limiter = None
    _validate_token_func = None

    def __init__(self, app=None, **kwargs):
        """Initialize dh-potluck."""
        self._app = app

        if app is not None:
            self.init_app(app, **kwargs)

    def init_app(self, app, **kwargs):
        self._app = app
        self._app_token = app.config['DH_POTLUCK_APP_TOKEN']
        self._structured_logging_enabled = app.config.get('STRUCTURED_LOGGING')
        self._rate_limiting_enabled = bool(app.config.get('RATELIMIT_ENABLED', 0))
        self._rate_limit = app.config.get('RATELIMIT_DEFAULT_PER_MINUTE', 1000)

        # Import function we use to authenticate incoming requests
        validate_func_name = app.config.get(
            'DH_POTLUCK_VALIDATE_TOKEN_FUNC', 'dh_potluck.auth.validate_token_using_api'
        )
        module_name, class_name = validate_func_name.rsplit('.', 1)
        self._validate_token_func = getattr(importlib.import_module(module_name), class_name)

        # Adjust the WSGI environ based on X-Forwarded- headers that proxies in front of the
        # application may set.
        app.wsgi_app = ProxyFix(app.wsgi_app, x_for=1, x_proto=1)

        # Disable Datadog tracing when testing
        tracer.configure(enabled=not app.config.get('TESTING', False))
        tracer.set_tags(
            {'dh_potluck.version': pkg_resources.get_distribution('dh-potluck').version}
        )

        self._configure_api_docs()
        self._limiter = self._configure_rate_limiting()
        self._app.before_request(self._process_request)
        self._app.after_request(self._process_response)
        self._register_error_handlers()

        if app.config.get('STRUCTURED_LOGGING'):
            self._configure_structured_logging()

        # Datadog Profiling - ddtrace 0.39.0 required
        if app.config.get('DD_PROFILING'):
            import ddtrace.profiling.auto  # noqa: F401

    def _get_rate_limit_key(self):
        """
        Extract a key from the request we can identify users by so we can rate limit appropriately.
        """
        if 'Authorization' in request.headers:
            return request.headers['Authorization'].split(' ')[1]
        else:
            ip = request.remote_addr
            if request.headers.getlist('X-Forwarded-For'):
                ip = request.headers.getlist('X-Forwarded-For')[0]
            return ip

    def _configure_api_docs(self):
        """
        Render ReDoc for DH API docs.
        """
        api_docs = Blueprint(
            'API Docs', __name__, url_prefix='/docs', template_folder='./templates'
        )

        @api_docs.route('')
        def render_docs():
            return render_template('dhdocs.html')

        self._app.register_blueprint(api_docs)

    def _configure_rate_limiting(self):
        limiter = Limiter(
            key_func=self._get_rate_limit_key,
            default_limits=[f'{self._rate_limit} per minute'],
            default_limits_per_method=False,
            headers_enabled=True,
            swallow_errors=True,
            enabled=self._rate_limiting_enabled,
        )

        @limiter.request_filter
        def app_token_whitelist():
            return request.headers.get('Authorization', '') == f'Application {self._app_token}'

        limiter.init_app(self._app)
        return limiter

    def _configure_structured_logging(self):
        # Configure root logger
        structure_logger(logging.getLogger())

        # Set all others
        for logger in [logging.getLogger(name) for name in logging.root.manager.loggerDict]:
            structure_logger(logger)

        # Allow for Celery logs to report log level to Datadog
        # TODO - remove this once Celery log configuration hooks are sorted
        patch_celery_get_logger()

    def _process_request(self):
        """
        Authenticate every incoming request.
        """
        # Allow any OPTIONS request so CORS works properly
        if request.method == 'OPTIONS':
            return

        # Add Domo session to Datadog traces if provided
        if 'domo-session' in request.headers:
            span = tracer.current_span()
            if span:
                span.set_tags({'dh_potluck.domo.session': request.headers.get('domo-session')})
        return get_user(self._app_token, self._validate_token_func)

    def _process_response(self, response):
        """
        Capture SQL queries made during request and log info about them.
        """
        if os.environ.get('FLASK_DEBUG', 'false') == 'true':
            message = get_database_queries_summary(self._app)
            self._app.logger.info(message)

        return response

    def _register_error_handlers(self):
        # Catch flask-smorest validation errors and return them in JSON format
        @self._app.errorhandler(HTTPStatus.UNPROCESSABLE_ENTITY)
        def handle_unprocessable_entity(error):
            add_request_params_to_trace()
            response = {
                'description': 'Input failed validation.',
                'errors': error.exc.messages,
            }
            return jsonify(response), HTTPStatus.BAD_REQUEST

        # Catch marshmallow validation errors and return them in JSON format
        @self._app.errorhandler(ValidationError)
        def handle_validation_error(error):
            add_request_params_to_trace()
            response = {
                'description': 'Input failed validation.',
                'errors': error.messages,
            }
            return jsonify(response), HTTPStatus.BAD_REQUEST

        # Catch SQLAlchemy IntegrityErrors (usually unique constraint violations) and return them
        # in JSON format. TODO: Right now we return the database error as-is to the client. This
        # should be expanded to parse the integrity error and try to build a more structured,
        # user-friendly message about the error.
        @self._app.errorhandler(IntegrityError)
        def handle_integrity_errors(error):
            add_request_params_to_trace()
            return (
                jsonify({'description': f'Database integrity error: {error.orig.args[1]}'}),
                HTTPStatus.BAD_REQUEST,
            )

        # Ensure all other Flask HTTP exceptions are returned in JSON format
        @self._app.errorhandler(HTTPException)
        def handle_flask_exceptions(error):
            add_request_params_to_trace()
            return jsonify({'description': error.description}), error.code

        # Add extra context to Datadog traces for server errors
        @self._app.errorhandler(HTTPStatus.INTERNAL_SERVER_ERROR)
        def handle_server_error(error):
            add_request_params_to_trace()
            error_response = (
                jsonify({'description': InternalServerError.description}),
                HTTPStatus.INTERNAL_SERVER_ERROR,
            )
            return error_response

        # Add extra context to Datadog traces for rate limited requests
        @self._app.errorhandler(HTTPStatus.TOO_MANY_REQUESTS)
        def handle_too_many_requests(error):
            if 'Authorization' in request.headers:
                key = self._get_rate_limit_key()
                cleaned_key = f'{key[0:4]}...{key[-4:-1]}'
            else:
                cleaned_key = request.remote_addr

            tracer.set_tags({'rate_limit_key': cleaned_key})

            return (
                jsonify(
                    {
                        'description': (
                            'The user has sent too many requests in a given amount of time.'
                        )
                    }
                ),
                HTTPStatus.TOO_MANY_REQUESTS,
            )

        @self._app.errorhandler(HTTPStatus.NOT_FOUND)
        def page_not_found(e):
            """Don't return 404 on OPTIONS calls"""
            if request and request.method == 'OPTIONS':
                return '', HTTPStatus.OK

            error_response = (
                jsonify({'description': e.description}),
                HTTPStatus.NOT_FOUND,
            )

            return error_response

        @self._app.errorhandler(PlatformConnectionError)
        def handle_platform_connection_error(err):
            return jsonify({'description': str(err)}), HTTPStatus.BAD_REQUEST

        @self._app.errorhandler(Exception)
        def handle_error(e):
            # Catch and log unhandled exceptions in JSON format
            if self._structured_logging_enabled:
                extra = {
                    'error.stack': traceback.format_exc(),
                    'error.kind': str(type(e)),
                    'logger.thread_name': threading.Thread.getName(threading.current_thread()),
                }
                self._app.logger.error(str(e), extra=extra)
            else:
                traceback.print_exc()

            add_request_params_to_trace()
            return (
                jsonify({'description': InternalServerError.description}),
                HTTPStatus.INTERNAL_SERVER_ERROR,
            )

    @staticmethod
    def role_required(*args, **kwargs):
        return role_required(*args, **kwargs)

    @property
    def current_user(self):
        return g.user

    @property
    def limiter(self):
        return self._limiter



from flask_sqlalchemy import get_debug_queries

from .utils.term import Bell, TextColors


def get_database_queries_summary(app):
    """
    Prepare a summary of the SQL queries executed during a request. Useful finding slow queries.
    """
    queries = get_debug_queries()
    message = f'\n{TextColors.LIGHTYELLOW}Database Queries Summary'
    slow_query_time = app.config.get('DH_SLOW_DB_QUERY_TIME', 0.5)
    too_many_queries = app.config.get('DH_TOO_MANY_SQL_QUERIES', 10)
    total_query_time = 0
    select_queries_count = 0
    slow_queries = []
    for query in queries:
        total_query_time += query.duration
        if query.statement.startswith('SELECT'):
            select_queries_count += 1
        if query.duration > slow_query_time:
            slow_queries.append(query)
    message += f'\n\t{TextColors.BLUE}Select queries: {TextColors.RESET}{select_queries_count}'

    if len(queries) - select_queries_count > 0:
        message += (
            f'\n\t{TextColors.BLUE}Other queries: '
            f'{TextColors.RESET}{len(queries) - select_queries_count}'
        )
    message += f'\n\t{TextColors.BLUE}Total queries: {TextColors.RESET}{len(queries)}'
    message += f'\n\t{TextColors.BLUE}Total query time: {TextColors.RESET}{total_query_time:.2f}s'

    if slow_queries:
        message += (
            f'\n\t{TextColors.LIGHTRED}{Bell.BELL}[⚠️  WARNING] Slow queries '
            f'(>{slow_query_time}s): {len(slow_queries)}'
        )

    if len(queries) > too_many_queries:
        message += (
            f'\n\t{TextColors.LIGHTRED}{Bell.BELL}[⚠️  WARNING] Too many queries '
            f'(max {too_many_queries} queries)'
        )

    return message



from .auth import ApplicationUser, UnauthenticatedUser
from .celery import signals as CelerySignals
from .celery.synchronization import Semaphore, SemaphoreLocked
from .decorators import retry
from .email import EmailService
from .environment import Environment
from .extension import DHPotluck
from .fields import EnumField
from .health_checks import HealthChecks
from .image_api import ImageApi
from .messaging import (
    IncomingMessageRouter,
    Message,
    MessageConsumer,
    MessageConsumerApp,
    MessageEnvelopeSchema,
    MessageHandler,
    MessageHandlerCallback,
    MessageProducer,
    message_handler,
)
from .platform_connection import (
    BadApiResponse,
    InvalidPlatformConnection,
    MissingPlatformConnection,
    PlatformConnection,
)

__all__ = [
    'DHPotluck',
    'EnumField',
    'ApplicationUser',
    'UnauthenticatedUser',
    'PlatformConnection',
    'BadApiResponse',
    'MissingPlatformConnection',
    'InvalidPlatformConnection',
    'CelerySignals',
    'HealthChecks',
    'retry',
    'Environment',
    'ImageApi',
    'IncomingMessageRouter',
    'Message',
    'MessageConsumer',
    'MessageConsumerApp',
    'MessageEnvelopeSchema',
    'MessageHandler',
    'MessageHandlerCallback',
    'MessageProducer',
    'message_handler',
    'EmailService',
    'Semaphore',
    'SemaphoreLocked',
]



from ddtrace import tracer

try:
    from celery.signals import task_prerun

    @task_prerun.connect
    def celer_task_prerun(task_id, task, args, **kwargs):
        # This will add the args specified in the celery task "dd_args" argument and add
        # them as tags to the current trace.

        # get the current span
        current_span = tracer.current_span()
        # set the host just once on the current span
        if current_span:
            if hasattr(task, 'dd_args'):
                for arg in task.dd_args:
                    if arg in task.request.kwargs:
                        current_span.set_tag(f'celery.task_args.{arg}', task.request.kwargs[arg])


except ModuleNotFoundError:
    pass



import json
import logging
import re
from urllib.parse import urlencode

import json_log_formatter
from boltons.iterutils import remap
from ddtrace import tracer
from flask import request

SENSITIVE_KEYS = re.compile(r'password|token|secret|key', flags=re.I)
MAX_BODY_SIZE = 50000
LOG_LEVEL_MAPPING = {
    'CRITICAL': 50,
    'ERROR': 40,
    'WARNING': 30,
    'INFO': 20,
    'DEBUG': 10,
}


def add_request_params_to_trace():
    """
    Add additional details about the current request (query string, body, request size) to
    the current Datadog trace. Useful for error handlers.
    """
    span = tracer.current_root_span()
    if not span:
        return

    # Log query string (if present) for all request methods
    query_params = request.args
    if query_params:
        clean = remap(query_params.copy(), visit=scrub_keys)
        span.set_tag('http.query_string', urlencode(clean))

    # Skip body logging if not POST, PATCH or PUT
    if request.method not in ['POST', 'PATCH', 'PUT']:
        return

    # Skip body logging if it's empty
    if not request.content_length:
        return

    span.set_tag('http.content_length', str(request.content_length))

    if request.content_length > MAX_BODY_SIZE:
        span.set_tag('http.body', 'Body too large, content could not be logged.')
        return

    # Try to parse body as JSON, and scrub sensitive values
    body = request.get_json(silent=True)
    if body:
        clean = remap(body, visit=scrub_keys)
        span.set_tag('http.body', json.dumps(clean))
    else:
        # If we can't parse as JSON, log the raw body
        body = request.get_data(as_text=True)
        span.set_tag('http.body', body)


def scrub_keys(path, key, value):
    if isinstance(key, str) and SENSITIVE_KEYS.search(key):
        return key, '-' * 5
    return key, value


def structure_logger(logger):
    def patch_level_func(level_func, level):
        def _patch_level_func(*args, **kwargs):
            extra = {**kwargs.get('extra', {}), **{'level': level}}
            return level_func(*args, **{**kwargs, **{'extra': extra}})

        return _patch_level_func

    def patch_log_func(log_func):
        reversed_mapping = dict([reversed(i) for i in LOG_LEVEL_MAPPING.items()])

        def _patch_log_func(level, *args, **kwargs):
            extra = {
                **kwargs.get('extra', {}),
                **{'level': reversed_mapping.get(level)},
            }
            return log_func(level, *args, **{**kwargs, **{'extra': extra}})

        return _patch_log_func

    # Insert level into json log for each level function invocation
    # ex. logger.info, logger.warning
    for level in LOG_LEVEL_MAPPING.keys():
        func = getattr(logger, level.lower())
        setattr(logger, level.lower(), patch_level_func(func, level))

    # Insert level into json log for direct log function call
    # ex logger.log
    func = getattr(logger, 'log')
    setattr(logger, 'log', patch_log_func(func))

    # Assign JSON formatter to handler
    json_handler = logging.StreamHandler()
    json_handler.setFormatter(json_log_formatter.JSONFormatter())
    logger.handlers = [json_handler]
    logger.propagate = False


def patch_celery_get_logger():
    # This has the opportunity of being a patch closer to the source (kombu)
    # Imports safely hidden behind structured_logging flag for now
    try:
        from celery.utils import log

        def _patch_func(name):
            l = log._get_logger(name)  # noqa: E741
            if logging.root not in (l, l.parent) and l is not log.base_logger:
                l = log._using_logger_parent(log.base_logger, l)  # noqa: E741
            structure_logger(l)
            return l

        log.get_logger = _patch_func
    except ModuleNotFoundError:
        return



import logging
from typing import Any, Dict

from marshmallow import Schema

from dh_potluck.messaging import IncomingMessageRouter, MessageHandler, MessageHandlerCallback

LOG = logging.getLogger(__name__)


_HANDLERS = {}


def message_handler(topic: str, schema: Schema):
    """
    Registers this function as a message handler
    :param str topic: Topic to handle
    :param Schema schema: Schema to use to deserialize the message payload
    """

    def decorator_register_message_handler(func: MessageHandlerCallback):
        _HANDLERS[topic] = MessageHandler(schema, func)
        return func

    return decorator_register_message_handler


class MessageConsumerApp(object):

    _router: IncomingMessageRouter

    def __init__(self, consumer_group: str, config_overrides: Dict[str, Any] = None):
        """
        :param str consumer_group: The kafka consumer group to use
        :param dict config_overrides: Any kafka configuration overrides
        """
        self._router = IncomingMessageRouter(_HANDLERS, consumer_group, config_overrides)

    def run(self):
        """
        Start consuming messages. On new messages, use the registered message_handler to handle it.
        :return: None
        """
        self._router.run()



from typing import Dict, List

from dh_potluck.email.mandrill_email_client import MandrillEmailClient


class EmailService(object):
    _email_client: MandrillEmailClient = None

    @classmethod
    def _get_email_client(cls):
        if not cls._email_client:
            cls._email_client = MandrillEmailClient()
        return cls._email_client

    @classmethod
    def send_email_template(
        cls,
        recipients: List[Dict],
        template_name: str,
        template_vars: List[Dict[str, str]],
        subject: str = None,
    ) -> None:
        try:
            cls._get_email_client().send_email_template(
                recipients, template_name, template_vars, subject=subject
            )
        except Exception as e:
            raise EmailServiceException(
                message=f'Error occurred while sending email template: {str(e)}'
            )

    @classmethod
    def send_email_html(cls, recipients: List[Dict], subject: str = None, html: str = None) -> None:
        try:
            cls._get_email_client().send_email_html(recipients, subject=subject, html=html)
        except Exception as e:
            raise EmailServiceException(
                message=f'Error occurred while sending html email: {str(e)}'
            )


class EmailServiceException(Exception):
    def __init__(self, message: str = 'Error occurred while sending email.') -> None:
        self.message = message
        super().__init__(self.message)



import logging
from datetime import datetime, timezone
from typing import Any, Callable, Dict, Union

from confluent_kafka import Producer
from confluent_kafka.cimpl import KafkaError, KafkaException
from confluent_kafka.cimpl import Message as KafkaMessage
from ddtrace import tracer
from flask import current_app
from marshmallow import Schema, ValidationError

from dh_potluck.messaging import MessageEnvelopeSchema

LOG = logging.getLogger(__name__)


class MessageProducer(object):

    _producer: Producer
    _message_schema: MessageEnvelopeSchema = MessageEnvelopeSchema()

    def __init__(self, config_overrides: Dict[str, Any] = None):
        """
        :param dict config_overrides: Any kafka configuration overrides
        """
        brokers: str = current_app.config['KAFKA_BROKERS_LIST']
        if brokers is None:
            raise RuntimeError(
                'Tried to instantiate a MessageProducer '
                'without setting the KAFKA_BROKERS_LIST env var'
            )
        config = {'bootstrap.servers': brokers}
        should_connect_ssl: bool = current_app.config['KAFKA_USE_SSL_CONNECTION']
        if should_connect_ssl:
            config['security.protocol'] = 'SSL'
        if config_overrides:
            config.update(config_overrides)
        self._producer = Producer(config)

    @staticmethod
    def _default_delivery_report(error: KafkaError, message: KafkaMessage) -> None:
        """Called once for each message produced to indicate delivery result.
        Triggered by poll() or flush()."""
        if error is not None:
            LOG.error(f'Message delivery failed: {error}')
        else:
            LOG.info(f'Message delivered to {message.topic()} [{message.partition()}]')

    def send_message(
        self,
        topic: str,
        message_type: str,
        payload: str,
        key: Union[str, bytes] = None,
        on_delivery: Callable[[KafkaError, KafkaMessage], None] = None,
        poll_timeout: float = 0,
    ) -> None:
        """
        Send a message

        :param str topic: Topic to send the message on
        :param str message_type: Message type
        :param str payload: Serialized message payload
        :param str|bytes key: (Optional) Message key, used to determine the partition the message
                              will go to
        :param func on_delivery: (Optional) Delivery report callback to call on successful or failed
                                 delivery
        :param float poll_timeout: Maximum time to block waiting for messages. (Seconds)
        :return: None
        :raises ValidationError: if the message doesn't conform to tge expected MessageSchema
        :raises BufferError: if the internal producer message queue is full
                             (``queue.buffering.max.messages`` exceeded)
        :raises KafkaException: for other errors, see exception code
        """
        LOG.info(f'Sending message on topic {topic}')
        body = {
            'message_type': message_type,
            'message_time': datetime.now(timezone.utc),
            'payload': payload,
        }
        body_str = self._serialize(body, self._message_schema)
        try:
            with tracer.trace('kafka.produce', resource=f'{topic} {message_type}'):
                self._producer.produce(
                    topic,
                    body_str.encode('utf-8'),
                    key=key,
                    on_delivery=on_delivery or self._default_delivery_report,
                )
                self._producer.poll(poll_timeout)
        except BufferError as e:
            LOG.error(f'Message buffer is full!: {e}')
            raise e
        except KafkaException as e:
            LOG.error(f'Failed to send message to broker: {e}')
            raise e

    def send_message_with_schema(
        self,
        topic: str,
        message_type: str,
        payload: dict,
        schema: Schema,
        key: Union[str, bytes] = None,
        on_delivery: Callable[[KafkaError, KafkaMessage], None] = None,
        poll_timeout: float = 0,
    ) -> None:
        """
        Send a message

        :param str topic: Topic to send the message on
        :param str message_type: Message type
        :param dict payload: Message payload
        :param Schema schema: Marshmallow schema to serialize payload
        :param str|bytes key: (Optional) Message key, used to determine the partition the message
                              will go to
        :param func on_delivery: (Optional) Delivery report callback to call on successful or failed
                                 delivery
        :param float poll_timeout: Maximum time to block waiting for messages. (Seconds)
        :return: None
        :raises ValidationError: if the message doesn't conform to tge expected MessageSchema
        :raises BufferError: if the internal producer message queue is full
                             (``queue.buffering.max.messages`` exceeded)
        :raises KafkaException: for other errors, see exception code
        """
        payload_str = self._serialize(payload, schema)

        self.send_message(
            topic=topic,
            message_type=message_type,
            payload=payload_str,
            key=key,
            on_delivery=on_delivery,
            poll_timeout=poll_timeout,
        )

    def _serialize(self, payload: Dict, schema: Schema) -> str:
        try:
            payload_str: str = schema.dumps(payload)
            schema.loads(payload_str)
        except ValidationError as e:
            LOG.error(f'Validation error(s) while sending message: {e}')
            raise e
        return payload_str

    def shutdown(self) -> None:
        """
        Call when service is shutting down
        :return: None
        """
        LOG.info('Producer shutting down...')
        self._producer.flush()
        LOG.info('Producer shut down')

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()



from marshmallow import EXCLUDE, Schema, fields


class MessageEnvelopeSchema(Schema):
    message_type = fields.String(required=True)
    message_time = fields.DateTime(required=True)
    payload = fields.String(required=True)

    class Meta:
        unknown = EXCLUDE



from http import HTTPStatus

from flask import Blueprint, abort

healthz = Blueprint('healthz', __name__)


class HealthChecks:
    callback = lambda: True  # noqa: E731

    def init_app(self, app, callback=None):
        if callback:
            HealthChecks.callback = callback
        app.register_blueprint(healthz, url_prefix='/healthz')


@healthz.route('/liveness')
def liveness():
    return 'OK'


@healthz.route('/readiness')
def readiness():
    try:
        if not HealthChecks.callback():
            abort(HTTPStatus.SERVICE_UNAVAILABLE, 'Pod unready to service requests.')
    except Exception as e:
        abort(HTTPStatus.SERVICE_UNAVAILABLE, str(e))
    return 'OK'



import base64
from enum import Enum

from flask import current_app

DEFAULT_PROD_CDN_URLS = ['https://dashhudson-static.s3.amazonaws.com', 'https://cdn.dashhudson.com']
DEFAULT_PROD_IMAGE_API_URL = 'https://images.dashhudson.com'
DEFAULT_DEV_IMAGE_API_URL = 'https://images.dhdev.io'


class ImageApi:
    class FitType(Enum):
        INSIDE = 'inside'
        COVER = 'cover'

    @classmethod
    def build_url(
        cls,
        url_orig: str,
        w: int = None,
        h: int = None,
        fit: FitType = None,
        download: bool = False,
        extension: str = 'jpg',
    ):
        """
        Returns a formatted URL to the Image API service
        :param url_orig: String, full URL to Dash Hudson image
            e.g. https://dashhudson-dev.s3.amazonaws.com/images/items/1532976128.41429521549.jpeg
        :param w: Int, requested width
        :param h: Int, requested height
        :param fit: FitType, how the image should be fit into the dimensions
            INSIDE: fit the image inside the bounds, maintaining aspect ratio
            COVER: fit the image to the size of the box, using a center crop if necessary
        :param download: Bool, flag that's sent to the image API to receive a download header,
            default False
        :param extension: String, image file extension, default 'jpg'
        """
        host = cls._get_image_api_host_for_url(url_orig)

        url = url_orig.encode('utf-8')
        url_bytes = base64.urlsafe_b64encode(url)
        encoded_path = str(url_bytes, 'utf-8')

        params = []
        if w is not None:
            params.append(f'w={w}')
        if h is not None:
            params.append(f'h={h}')
        if fit is not None:
            params.append(f'fit={fit.value}')
        if download:
            params.append('download=true')

        param_string = f'?{"&".join(params)}' if len(params) > 0 else ''

        return f'{host}/{encoded_path}.{extension}{param_string}'

    @classmethod
    def _get_image_api_host_for_url(cls, url_orig):
        prod_cdn_urls = current_app.config.get(
            'DH_POTLUCK_PROD_IMAGE_CDN_URLS', DEFAULT_PROD_CDN_URLS
        )
        for cdn_url in prod_cdn_urls:
            if url_orig.startswith(cdn_url):
                host = current_app.config.get(
                    'DH_POTLUCK_PROD_IMAGE_API_URL', DEFAULT_PROD_IMAGE_API_URL
                )
                break
        else:
            host = current_app.config.get('DH_POTLUCK_DEV_IMAGE_API_URL', DEFAULT_DEV_IMAGE_API_URL)
        return host



from marshmallow.fields import Field


class EnumField(Field):
    """
    This replacement for EnumField is used to resolve an issue where the default marshmallow enum
    did not work to properly serialize allowed values and the property, missing, for the API spec.
    """

    default_error_messages = {
        'invalid': 'Invalid enum value {input}',
    }

    def __init__(self, enum_type, *args, **kwargs):
        self.enum = enum_type

        super().__init__(*args, **kwargs)

        # Detect type of enum and make it available to apispec
        values = [e.value for e in self.enum if e.value is not None]
        if all(isinstance(v, int) for v in values):
            self.metadata['type'] = 'integer'
        elif all(isinstance(v, (float, int)) for v in values):
            self.metadata['type'] = 'number'
        elif all(isinstance(v, bool) for v in values):
            self.metadata['type'] = 'boolean'
        elif all(isinstance(v, str) for v in values):
            self.metadata['type'] = 'string'

        # Ensure all enum values are made available to apispec
        self.metadata['enum'] = sorted([e.value for e in self.enum])

    # These template methods are not invoked when constructing the apispec
    # but they are used by the routes
    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return None
        return value.value

    def _deserialize(self, value, attr, data, **kwargs):
        if value is None:
            return None
        try:
            return self.enum(value)
        except ValueError as error:
            raise self.make_error('invalid', input=value, value=value) from error

    def make_error(self, key, **kwargs):
        kwargs['values'] = ', '.join([str(mem.value) for mem in self.enum])
        return super().make_error(key, **kwargs)
